from __future__ import annotations
from typing import (Optional, Tuple, List, Callable, Any)
from ...fable_library.array import (map, filter, try_find as try_find_1, map_indexed)
from ...fable_library.int32 import try_parse
from ...fable_library.option import (map as map_1, bind)
from ...fable_library.reflection import (TypeInfo, string_type, option_type, int32_type, record_type)
from ...fable_library.reg_exp import replace
from ...fable_library.string import substring
from ...fable_library.types import (Record, FSharpRef)
from ...fable_library.util import equals
from ...fs_hafas_python.context import (Context, Profile__get_parseBitmask)
from ...fs_hafas_python.extra_types import IndexMap_2
from ...fs_hafas_python.lib.transformations import (Coordinate_toFloat, Default_Location, Default_Stop, Default_Station)
from ...fs_hafas_python.types_hafas_client import (Location, Line, Stop, Station)
from ...fs_hafas_python.types_raw_hafas_client import (RawLoc, RawCrd, RawCommon)
from .common import (to_option, map_index_array)

def expr_428() -> TypeInfo:
    return record_type("FsHafas.Parser.Location.Lid", [], Lid, lambda: [("L", option_type(string_type)), ("O", option_type(string_type)), ("X", option_type(int32_type)), ("Y", option_type(int32_type))])


class Lid(Record):
    def __init__(self, L: Optional[str], O: Optional[str], X: Optional[int], Y: Optional[int]) -> None:
        super().__init__()
        self.L = L
        self.O = O
        self.X = X
        self.Y = Y
    

Lid_reflection = expr_428

def parse_lid(lid: Optional[str]=None) -> Lid:
    if lid is None:
        return Lid(None, None, None, None)
    
    else: 
        lid_1 : str = lid
        def mapping(s_1: str, lid: Optional[str]=lid) -> Tuple[str, str]:
            return (s_1[0], substring(s_1, 2))
        
        def predicate(s: str, lid: Optional[str]=lid) -> bool:
            return len(s) > 3
        
        tuples : List[Tuple[str, str]] = map(mapping, filter(predicate, lid_1.split("@")), None)
        def try_find(c: str, lid: Optional[str]=lid) -> Optional[str]:
            def mapping_1(tupled_arg_1: Tuple[str, str], c: str=c) -> str:
                return tupled_arg_1[1]
            
            def predicate_1(tupled_arg: Tuple[str, str], c: str=c) -> bool:
                return tupled_arg[0] == c
            
            return map_1(mapping_1, try_find_1(predicate_1, tuples))
        
        try_find : Callable[[str], Optional[str]] = try_find
        def try_parse_int(s_2: str, lid: Optional[str]=lid) -> Optional[int]:
            match_value : Tuple[bool, int]
            out_arg : int = 0
            def arrow_429(s_2: str=s_2) -> int:
                return out_arg
            
            def arrow_430(v_2: int, s_2: str=s_2) -> None:
                nonlocal out_arg
                out_arg = v_2 or 0
            
            match_value = (try_parse(s_2, 511, False, 32, FSharpRef(arrow_429, arrow_430)), out_arg)
            if match_value[0]:
                return match_value[1]
            
            else: 
                return None
            
        
        try_parse_int : Callable[[str], Optional[int]] = try_parse_int
        return Lid(try_find("L"), try_find("O"), bind(try_parse_int, try_find("X")), bind(try_parse_int, try_find("Y")))
    


def remove_leading_zeros(s: str) -> str:
    return replace(s, "^0+", "")


def parse_location_phase1(ctx: Context, i: int, locl: List[RawLoc]) -> Tuple[RawLoc, Any]:
    l : RawLoc = locl[i]
    lid : Lid = parse_lid(l.lid)
    def arrow_431(s: str, ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> str:
        return remove_leading_zeros(s)
    
    def arrow_432(ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> Optional[str]:
        match_value : Tuple[Optional[str], Optional[str]] = (l.ext_id, lid.L)
        return match_value[0] if (match_value[0] is not None) else (match_value[1] if (match_value[1] is not None) else None)
    
    id : Optional[str] = map_1(arrow_431, arrow_432())
    pattern_input : Tuple[Optional[float], Optional[float]]
    match_value_1 : Optional[RawCrd] = l.crd
    (pattern_matching_result, crd_1) = (None, None)
    if match_value_1 is not None:
        def arrow_433(ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> bool:
            crd : RawCrd = match_value_1
            return (crd.y > 0) if (crd.x > 0) else False
        
        if arrow_433():
            pattern_matching_result = 0
            crd_1 = match_value_1
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        pattern_input = (Coordinate_toFloat(crd_1.x), Coordinate_toFloat(crd_1.y))
    
    elif pattern_matching_result == 1:
        match_value_2 : Tuple[Optional[int], Optional[int]] = (lid.X, lid.Y)
        (pattern_matching_result_1, X, Y) = (None, None, None)
        if match_value_2[0] is not None:
            if match_value_2[1] is not None:
                pattern_matching_result_1 = 0
                X = match_value_2[0]
                Y = match_value_2[1]
            
            else: 
                pattern_matching_result_1 = 1
            
        
        else: 
            pattern_matching_result_1 = 1
        
        if pattern_matching_result_1 == 0:
            pattern_input = (Coordinate_toFloat(X), Coordinate_toFloat(Y))
        
        elif pattern_matching_result_1 == 1:
            pattern_input = (None, None)
        
    
    lon : Optional[float] = pattern_input[0]
    lat : Optional[float] = pattern_input[1]
    def binder(d: int, ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> Optional[int]:
        if d > 0:
            return d
        
        else: 
            return None
        
    
    distance : Optional[int] = bind(binder, l.dist)
    match_value_3 : Optional[str] = l.type
    if match_value_3 is None:
        return (l, Location(Default_Location.type, id, l.name, Default_Location.poi, Default_Location.address, lon, lat, Default_Location.altitude, distance))
    
    else: 
        type : str = match_value_3
        if type == "S":
            def mapping(_arg1: float, ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> Location:
                return Location(Default_Location.type, id, Default_Location.name, Default_Location.poi, Default_Location.address, lon, lat, Default_Location.altitude, Default_Location.distance)
            
            def mapping_1(p_cls: int, ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> IndexMap_2[str, bool]:
                return Profile__get_parseBitmask(ctx.profile)(ctx)(p_cls)
            
            def get_target_array(_arg2: RawCommon, ctx: Context=ctx, i: int=i, locl: List[RawLoc]=locl) -> Optional[List[Line]]:
                return ctx.common.lines
            
            return (l, Stop(Default_Stop.type, id, l.name, map_1(mapping, lon), Default_Stop.station, map_1(mapping_1, l.p_cls), to_option(map_index_array(ctx.res.common, get_target_array, l.p_ref_l)) if ctx.opt.lines_of_stops else None, Default_Stop.is_meta, Default_Stop.reisezentrum_opening_hours, Default_Stop.ids, Default_Stop.load_factor, Default_Stop.entrances, Default_Stop.transit_authority, distance))
        
        else: 
            return (l, Location(Default_Location.type, id, l.name if (type != "A") else None, Default_Location.poi, l.name if (type == "A") else None, lon, lat, Default_Location.altitude, distance))
        
    


def parse_location_phase2(i: int, l: RawLoc, locations: List[Tuple[RawLoc, Any]], common_locations: List[Any]) -> Stop:
    station : Optional[Station]
    match_value : Optional[int] = l.m_mast_loc_x
    (pattern_matching_result, m_mast_loc_x_1) = (None, None)
    if match_value is not None:
        if match_value < len(locations):
            pattern_matching_result = 0
            m_mast_loc_x_1 = match_value
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        match_value_1 : Tuple[RawLoc, Any] = locations[m_mast_loc_x_1]
        if isinstance(match_value_1[1], Stop):
            stop : Stop = match_value_1[1]
            station = Station(Default_Station.type, stop.id, stop.name, Default_Station.station, stop.location, stop.products, stop.lines, Default_Station.is_meta, Default_Station.regions, Default_Station.facilities, Default_Station.reisezentrum_opening_hours, Default_Station.stops, Default_Station.entrances, Default_Station.transit_authority, Default_Station.distance)
        
        else: 
            station = None
        
    
    elif pattern_matching_result == 1:
        (pattern_matching_result_1, m_mast_loc_x_3) = (None, None)
        if match_value is not None:
            if match_value < len(common_locations):
                pattern_matching_result_1 = 0
                m_mast_loc_x_3 = match_value
            
            else: 
                pattern_matching_result_1 = 1
            
        
        else: 
            pattern_matching_result_1 = 1
        
        if pattern_matching_result_1 == 0:
            match_value_2 : Any = common_locations[m_mast_loc_x_3]
            if isinstance(match_value_2, Stop):
                stop_1 : Stop = match_value_2
                station = Station(Default_Station.type, stop_1.id, stop_1.name, Default_Station.station, stop_1.location, stop_1.products, stop_1.lines, Default_Station.is_meta, Default_Station.regions, Default_Station.facilities, Default_Station.reisezentrum_opening_hours, Default_Station.stops, Default_Station.entrances, Default_Station.transit_authority, Default_Station.distance)
            
            else: 
                station = None
            
        
        elif pattern_matching_result_1 == 1:
            station = None
        
    
    match_value_3 : Tuple[RawLoc, Any] = locations[i]
    if isinstance(match_value_3[1], Stop):
        stop_2 : Stop = match_value_3[1]
        return Stop(stop_2.type, stop_2.id, stop_2.name, stop_2.location, station, stop_2.products, stop_2.lines, stop_2.is_meta, stop_2.reisezentrum_opening_hours, stop_2.ids, stop_2.load_factor, stop_2.entrances, stop_2.transit_authority, stop_2.distance)
    
    else: 
        return match_value_3[1]
    


def parse_locations(ctx: Context, loc_l: List[RawLoc]) -> List[Any]:
    def mapping(i: int, _arg1: RawLoc, ctx: Context=ctx, loc_l: List[RawLoc]=loc_l) -> Tuple[RawLoc, Any]:
        return parse_location_phase1(ctx, i, loc_l)
    
    locations : List[Tuple[RawLoc, Any]] = map_indexed(mapping, loc_l, None)
    def predicate(u3: Any=None, ctx: Context=ctx, loc_l: List[RawLoc]=loc_l) -> bool:
        (pattern_matching_result,) = (None,)
        if isinstance(u3, Location):
            def arrow_434(u3: Any=u3) -> bool:
                l_1 : Location = u3
                return True if equals(l_1.latitude, None) else equals(l_1.longitude, None)
            
            if arrow_434():
                pattern_matching_result = 0
            
            else: 
                pattern_matching_result = 1
            
        
        else: 
            pattern_matching_result = 1
        
        if pattern_matching_result == 0:
            return False
        
        elif pattern_matching_result == 1:
            return True
        
    
    def mapping_1(i_1: int, tupled_arg: Tuple[RawLoc, Any], ctx: Context=ctx, loc_l: List[RawLoc]=loc_l) -> Any:
        return parse_location_phase2(i_1, tupled_arg[0], locations, ctx.common.locations)
    
    return filter(predicate, map_indexed(mapping_1, locations, None))





from __future__ import annotations
from abc import abstractmethod
from typing import (MutableSequence, Optional, List, Protocol, Any)
from ..fable_library.reflection import (TypeInfo, string_type, int32_type, array_type, bool_type, record_type, option_type, float64_type, obj_type, class_type)
from ..fable_library.types import Record
from .extra_types import (IndexMap_2, IndexMap_2_reflection, Icon, Icon_reflection)

def expr_240() -> TypeInfo:
    return record_type("FsHafas.Client.ProductType", [], ProductType, lambda: [("id", string_type), ("mode", string_type), ("name", string_type), ("short", string_type), ("bitmasks", array_type(int32_type)), ("default", bool_type)])


class ProductType(Record):
    def __init__(self, id: str, mode: str, name: str, short: str, bitmasks: MutableSequence[int], default: bool) -> None:
        super().__init__()
        self.id = id
        self.mode = mode
        self.name = name
        self.short = short
        self.bitmasks = bitmasks
        self.default = default
    

ProductType_reflection = expr_240

class Profile(Protocol):
    @property
    @abstractmethod
    def endpoint(self) -> str:
        ...
    
    @property
    @abstractmethod
    def journeys_from_trip(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def journeys_walking_speed(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def lines(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def locale(self) -> str:
        ...
    
    @property
    @abstractmethod
    def products(self) -> List[ProductType]:
        ...
    
    @property
    @abstractmethod
    def radar(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def reachable_from(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def refresh_journey(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def remarks(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def remarks_get_polyline(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def timezone(self) -> str:
        ...
    
    @property
    @abstractmethod
    def trip(self) -> Optional[bool]:
        ...
    
    @property
    @abstractmethod
    def trips_by_name(self) -> Optional[bool]:
        ...
    

def expr_241() -> TypeInfo:
    return record_type("FsHafas.Client.Location", [], Location, lambda: [("type", option_type(string_type)), ("id", option_type(string_type)), ("name", option_type(string_type)), ("poi", option_type(bool_type)), ("address", option_type(string_type)), ("longitude", option_type(float64_type)), ("latitude", option_type(float64_type)), ("altitude", option_type(float64_type)), ("distance", option_type(int32_type))])


class Location(Record):
    def __init__(self, type: Optional[str], id: Optional[str], name: Optional[str], poi: Optional[bool], address: Optional[str], longitude: Optional[float], latitude: Optional[float], altitude: Optional[float], distance: Optional[int]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.name = name
        self.poi = poi
        self.address = address
        self.longitude = longitude
        self.latitude = latitude
        self.altitude = altitude
        self.distance = distance
    

Location_reflection = expr_241

def expr_242() -> TypeInfo:
    return record_type("FsHafas.Client.ReisezentrumOpeningHours", [], ReisezentrumOpeningHours, lambda: [("Mo", option_type(string_type)), ("Di", option_type(string_type)), ("Mi", option_type(string_type)), ("Do", option_type(string_type)), ("Fr", option_type(string_type)), ("Sa", option_type(string_type)), ("So", option_type(string_type))])


class ReisezentrumOpeningHours(Record):
    def __init__(self, Mo: Optional[str], Di: Optional[str], Mi: Optional[str], Do: Optional[str], Fr: Optional[str], Sa: Optional[str], So: Optional[str]) -> None:
        super().__init__()
        self.Mo = Mo
        self.Di = Di
        self.Mi = Mi
        self.Do = Do
        self.Fr = Fr
        self.Sa = Sa
        self.So = So
    

ReisezentrumOpeningHours_reflection = expr_242

def expr_243() -> TypeInfo:
    return record_type("FsHafas.Client.Station", [], Station, lambda: [("type", option_type(string_type)), ("id", option_type(string_type)), ("name", option_type(string_type)), ("station", option_type(Station_reflection())), ("location", option_type(Location_reflection())), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("lines", option_type(array_type(Line_reflection()))), ("is_meta", option_type(bool_type)), ("regions", option_type(array_type(string_type))), ("facilities", option_type(IndexMap_2_reflection(string_type, string_type))), ("reisezentrum_opening_hours", option_type(ReisezentrumOpeningHours_reflection())), ("stops", option_type(array_type(obj_type))), ("entrances", option_type(array_type(Location_reflection()))), ("transit_authority", option_type(string_type)), ("distance", option_type(int32_type))])


class Station(Record):
    def __init__(self, type: Optional[str], id: Optional[str], name: Optional[str], station: Optional[Station], location: Optional[Location], products: Optional[IndexMap_2[str, bool]], lines: Optional[List[Line]], is_meta: Optional[bool], regions: Optional[List[str]], facilities: Optional[IndexMap_2[str, str]], reisezentrum_opening_hours: Optional[ReisezentrumOpeningHours], stops: Optional[List[Any]], entrances: Optional[List[Location]], transit_authority: Optional[str], distance: Optional[int]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.name = name
        self.station = station
        self.location = location
        self.products = products
        self.lines = lines
        self.is_meta = is_meta
        self.regions = regions
        self.facilities = facilities
        self.reisezentrum_opening_hours = reisezentrum_opening_hours
        self.stops = stops
        self.entrances = entrances
        self.transit_authority = transit_authority
        self.distance = distance
    

Station_reflection = expr_243

def expr_244() -> TypeInfo:
    return record_type("FsHafas.Client.Stop", [], Stop, lambda: [("type", option_type(string_type)), ("id", option_type(string_type)), ("name", option_type(string_type)), ("location", option_type(Location_reflection())), ("station", option_type(Station_reflection())), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("lines", option_type(array_type(Line_reflection()))), ("is_meta", option_type(bool_type)), ("reisezentrum_opening_hours", option_type(ReisezentrumOpeningHours_reflection())), ("ids", option_type(IndexMap_2_reflection(string_type, string_type))), ("load_factor", option_type(string_type)), ("entrances", option_type(array_type(Location_reflection()))), ("transit_authority", option_type(string_type)), ("distance", option_type(int32_type))])


class Stop(Record):
    def __init__(self, type: Optional[str], id: Optional[str], name: Optional[str], location: Optional[Location], station: Optional[Station], products: Optional[IndexMap_2[str, bool]], lines: Optional[List[Line]], is_meta: Optional[bool], reisezentrum_opening_hours: Optional[ReisezentrumOpeningHours], ids: Optional[IndexMap_2[str, str]], load_factor: Optional[str], entrances: Optional[List[Location]], transit_authority: Optional[str], distance: Optional[int]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.name = name
        self.location = location
        self.station = station
        self.products = products
        self.lines = lines
        self.is_meta = is_meta
        self.reisezentrum_opening_hours = reisezentrum_opening_hours
        self.ids = ids
        self.load_factor = load_factor
        self.entrances = entrances
        self.transit_authority = transit_authority
        self.distance = distance
    

Stop_reflection = expr_244

def expr_245() -> TypeInfo:
    return record_type("FsHafas.Client.Region", [], Region, lambda: [("type", option_type(string_type)), ("id", string_type), ("name", string_type), ("stations", array_type(string_type))])


class Region(Record):
    def __init__(self, type: Optional[str], id: str, name: str, stations: List[str]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.name = name
        self.stations = stations
    

Region_reflection = expr_245

def expr_246() -> TypeInfo:
    return record_type("FsHafas.Client.Line", [], Line, lambda: [("type", option_type(string_type)), ("id", option_type(string_type)), ("name", option_type(string_type)), ("admin_code", option_type(string_type)), ("fahrt_nr", option_type(string_type)), ("additional_name", option_type(string_type)), ("product", option_type(string_type)), ("public", option_type(bool_type)), ("mode", option_type(string_type)), ("routes", option_type(array_type(string_type))), ("operator", option_type(Operator_reflection())), ("express", option_type(bool_type)), ("metro", option_type(bool_type)), ("night", option_type(bool_type)), ("nr", option_type(int32_type)), ("symbol", option_type(string_type)), ("directions", option_type(array_type(string_type))), ("product_name", option_type(string_type))])


class Line(Record):
    def __init__(self, type: Optional[str], id: Optional[str], name: Optional[str], admin_code: Optional[str], fahrt_nr: Optional[str], additional_name: Optional[str], product: Optional[str], public: Optional[bool], mode: Optional[str], routes: Optional[List[str]], operator: Optional[Operator], express: Optional[bool], metro: Optional[bool], night: Optional[bool], nr: Optional[int], symbol: Optional[str], directions: Optional[List[str]], product_name: Optional[str]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.name = name
        self.admin_code = admin_code
        self.fahrt_nr = fahrt_nr
        self.additional_name = additional_name
        self.product = product
        self.public = public
        self.mode = mode
        self.routes = routes
        self.operator = operator
        self.express = express
        self.metro = metro
        self.night = night
        self.nr = nr
        self.symbol = symbol
        self.directions = directions
        self.product_name = product_name
    

Line_reflection = expr_246

def expr_247() -> TypeInfo:
    return record_type("FsHafas.Client.Route", [], Route, lambda: [("type", option_type(string_type)), ("id", string_type), ("line", string_type), ("mode", string_type), ("stops", array_type(string_type))])


class Route(Record):
    def __init__(self, type: Optional[str], id: str, line: str, mode: str, stops: List[str]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.line = line
        self.mode = mode
        self.stops = stops
    

Route_reflection = expr_247

def expr_248() -> TypeInfo:
    return record_type("FsHafas.Client.Cycle", [], Cycle, lambda: [("min", option_type(int32_type)), ("max", option_type(int32_type)), ("nr", option_type(int32_type))])


class Cycle(Record):
    def __init__(self, min: Optional[int], max: Optional[int], nr: Optional[int]) -> None:
        super().__init__()
        self.min = min
        self.max = max
        self.nr = nr
    

Cycle_reflection = expr_248

def expr_249() -> TypeInfo:
    return record_type("FsHafas.Client.ArrivalDeparture", [], ArrivalDeparture, lambda: [("arrival", option_type(float64_type)), ("departure", option_type(float64_type))])


class ArrivalDeparture(Record):
    def __init__(self, arrival: Optional[float], departure: Optional[float]) -> None:
        super().__init__()
        self.arrival = arrival
        self.departure = departure
    

ArrivalDeparture_reflection = expr_249

def expr_250() -> TypeInfo:
    return record_type("FsHafas.Client.Schedule", [], Schedule, lambda: [("type", option_type(string_type)), ("id", string_type), ("route", string_type), ("mode", string_type), ("sequence", array_type(ArrivalDeparture_reflection())), ("starts", array_type(string_type))])


class Schedule(Record):
    def __init__(self, type: Optional[str], id: str, route: str, mode: str, sequence: List[ArrivalDeparture], starts: List[str]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.route = route
        self.mode = mode
        self.sequence = sequence
        self.starts = starts
    

Schedule_reflection = expr_250

def expr_251() -> TypeInfo:
    return record_type("FsHafas.Client.Operator", [], Operator, lambda: [("type", option_type(string_type)), ("id", string_type), ("name", string_type)])


class Operator(Record):
    def __init__(self, type: Optional[str], id: str, name: str) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.name = name
    

Operator_reflection = expr_251

def expr_252() -> TypeInfo:
    return record_type("FsHafas.Client.Hint", [], Hint, lambda: [("type", option_type(string_type)), ("code", option_type(string_type)), ("summary", option_type(string_type)), ("text", string_type), ("trip_id", option_type(string_type))])


class Hint(Record):
    def __init__(self, type: Optional[str], code: Optional[str], summary: Optional[str], text: str, trip_id: Optional[str]) -> None:
        super().__init__()
        self.type = type
        self.code = code
        self.summary = summary
        self.text = text
        self.trip_id = trip_id
    

Hint_reflection = expr_252

def expr_253() -> TypeInfo:
    return record_type("FsHafas.Client.Status", [], Status, lambda: [("type", option_type(string_type)), ("code", option_type(string_type)), ("summary", option_type(string_type)), ("text", string_type), ("trip_id", option_type(string_type))])


class Status(Record):
    def __init__(self, type: Optional[str], code: Optional[str], summary: Optional[str], text: str, trip_id: Optional[str]) -> None:
        super().__init__()
        self.type = type
        self.code = code
        self.summary = summary
        self.text = text
        self.trip_id = trip_id
    

Status_reflection = expr_253

def expr_254() -> TypeInfo:
    return record_type("FsHafas.Client.IcoCrd", [], IcoCrd, lambda: [("x", int32_type), ("y", int32_type), ("type", option_type(string_type))])


class IcoCrd(Record):
    def __init__(self, x: int, y: int, type: Optional[str]) -> None:
        super().__init__()
        self.x = x or 0
        self.y = y or 0
        self.type = type
    

IcoCrd_reflection = expr_254

def expr_255() -> TypeInfo:
    return record_type("FsHafas.Client.Edge", [], Edge, lambda: [("from_loc", option_type(obj_type)), ("to_loc", option_type(obj_type)), ("icon", option_type(Icon_reflection())), ("dir", option_type(int32_type)), ("ico_crd", option_type(IcoCrd_reflection()))])


class Edge(Record):
    def __init__(self, from_loc: Optional[Any], to_loc: Optional[Any], icon: Optional[Icon], dir: Optional[int], ico_crd: Optional[IcoCrd]) -> None:
        super().__init__()
        self.from_loc = from_loc
        self.to_loc = to_loc
        self.icon = icon
        self.dir = dir
        self.ico_crd = ico_crd
    

Edge_reflection = expr_255

def expr_256() -> TypeInfo:
    return record_type("FsHafas.Client.Event", [], Event, lambda: [("from_loc", option_type(obj_type)), ("to_loc", option_type(obj_type)), ("start", option_type(string_type)), ("end", option_type(string_type)), ("sections", option_type(array_type(string_type)))])


class Event(Record):
    def __init__(self, from_loc: Optional[Any], to_loc: Optional[Any], start: Optional[str], end: Optional[str], sections: Optional[List[str]]) -> None:
        super().__init__()
        self.from_loc = from_loc
        self.to_loc = to_loc
        self.start = start
        self.end = end
        self.sections = sections
    

Event_reflection = expr_256

def expr_257() -> TypeInfo:
    return record_type("FsHafas.Client.Warning", [], Warning, lambda: [("type", option_type(string_type)), ("id", option_type(string_type)), ("icon", option_type(Icon_reflection())), ("summary", option_type(string_type)), ("text", option_type(string_type)), ("category", option_type(string_type)), ("priority", option_type(int32_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("edges", option_type(array_type(Edge_reflection()))), ("events", option_type(array_type(Event_reflection()))), ("valid_from", option_type(string_type)), ("valid_until", option_type(string_type)), ("modified", option_type(string_type)), ("company", option_type(string_type)), ("categories", option_type(array_type(int32_type))), ("affected_lines", option_type(array_type(Line_reflection()))), ("from_stops", option_type(array_type(obj_type))), ("to_stops", option_type(array_type(obj_type)))])


class Warning(Record):
    def __init__(self, type: Optional[str], id: Optional[str], icon: Optional[Icon], summary: Optional[str], text: Optional[str], category: Optional[str], priority: Optional[int], products: Optional[IndexMap_2[str, bool]], edges: Optional[List[Edge]], events: Optional[List[Event]], valid_from: Optional[str], valid_until: Optional[str], modified: Optional[str], company: Optional[str], categories: Optional[MutableSequence[int]], affected_lines: Optional[List[Line]], from_stops: Optional[List[Any]], to_stops: Optional[List[Any]]) -> None:
        super().__init__()
        self.type = type
        self.id = id
        self.icon = icon
        self.summary = summary
        self.text = text
        self.category = category
        self.priority = priority
        self.products = products
        self.edges = edges
        self.events = events
        self.valid_from = valid_from
        self.valid_until = valid_until
        self.modified = modified
        self.company = company
        self.categories = categories
        self.affected_lines = affected_lines
        self.from_stops = from_stops
        self.to_stops = to_stops
    

Warning_reflection = expr_257

def expr_258() -> TypeInfo:
    return record_type("FsHafas.Client.Geometry", [], Geometry, lambda: [("type", option_type(string_type)), ("coordinates", array_type(float64_type))])


class Geometry(Record):
    def __init__(self, type: Optional[str], coordinates: MutableSequence[float]) -> None:
        super().__init__()
        self.type = type
        self.coordinates = coordinates
    

Geometry_reflection = expr_258

def expr_259() -> TypeInfo:
    return record_type("FsHafas.Client.Feature", [], Feature, lambda: [("type", option_type(string_type)), ("properties", obj_type), ("geometry", Geometry_reflection())])


class Feature(Record):
    def __init__(self, type: Optional[str], properties: Any, geometry: Geometry) -> None:
        super().__init__()
        self.type = type
        self.properties = properties
        self.geometry = geometry
    

Feature_reflection = expr_259

def expr_260() -> TypeInfo:
    return record_type("FsHafas.Client.FeatureCollection", [], FeatureCollection, lambda: [("type", option_type(string_type)), ("features", array_type(Feature_reflection()))])


class FeatureCollection(Record):
    def __init__(self, type: Optional[str], features: List[Feature]) -> None:
        super().__init__()
        self.type = type
        self.features = features
    

FeatureCollection_reflection = expr_260

def expr_261() -> TypeInfo:
    return record_type("FsHafas.Client.StopOver", [], StopOver, lambda: [("stop", option_type(obj_type)), ("departure", option_type(string_type)), ("departure_delay", option_type(int32_type)), ("prognosed_departure", option_type(string_type)), ("planned_departure", option_type(string_type)), ("departure_platform", option_type(string_type)), ("prognosed_departure_platform", option_type(string_type)), ("planned_departure_platform", option_type(string_type)), ("arrival", option_type(string_type)), ("arrival_delay", option_type(int32_type)), ("prognosed_arrival", option_type(string_type)), ("planned_arrival", option_type(string_type)), ("arrival_platform", option_type(string_type)), ("prognosed_arrival_platform", option_type(string_type)), ("planned_arrival_platform", option_type(string_type)), ("remarks", option_type(array_type(obj_type))), ("pass_by", option_type(bool_type)), ("cancelled", option_type(bool_type))])


class StopOver(Record):
    def __init__(self, stop: Optional[Any], departure: Optional[str], departure_delay: Optional[int], prognosed_departure: Optional[str], planned_departure: Optional[str], departure_platform: Optional[str], prognosed_departure_platform: Optional[str], planned_departure_platform: Optional[str], arrival: Optional[str], arrival_delay: Optional[int], prognosed_arrival: Optional[str], planned_arrival: Optional[str], arrival_platform: Optional[str], prognosed_arrival_platform: Optional[str], planned_arrival_platform: Optional[str], remarks: Optional[List[Any]], pass_by: Optional[bool], cancelled: Optional[bool]) -> None:
        super().__init__()
        self.stop = stop
        self.departure = departure
        self.departure_delay = departure_delay
        self.prognosed_departure = prognosed_departure
        self.planned_departure = planned_departure
        self.departure_platform = departure_platform
        self.prognosed_departure_platform = prognosed_departure_platform
        self.planned_departure_platform = planned_departure_platform
        self.arrival = arrival
        self.arrival_delay = arrival_delay
        self.prognosed_arrival = prognosed_arrival
        self.planned_arrival = planned_arrival
        self.arrival_platform = arrival_platform
        self.prognosed_arrival_platform = prognosed_arrival_platform
        self.planned_arrival_platform = planned_arrival_platform
        self.remarks = remarks
        self.pass_by = pass_by
        self.cancelled = cancelled
    

StopOver_reflection = expr_261

def expr_262() -> TypeInfo:
    return record_type("FsHafas.Client.Trip", [], Trip, lambda: [("id", string_type), ("origin", option_type(obj_type)), ("destination", option_type(obj_type)), ("departure", option_type(string_type)), ("planned_departure", option_type(string_type)), ("prognosed_arrival", option_type(string_type)), ("departure_delay", option_type(int32_type)), ("departure_platform", option_type(string_type)), ("prognosed_departure_platform", option_type(string_type)), ("planned_departure_platform", option_type(string_type)), ("arrival", option_type(string_type)), ("planned_arrival", option_type(string_type)), ("prognosed_departure", option_type(string_type)), ("arrival_delay", option_type(int32_type)), ("arrival_platform", option_type(string_type)), ("prognosed_arrival_platform", option_type(string_type)), ("planned_arrival_platform", option_type(string_type)), ("stopovers", option_type(array_type(StopOver_reflection()))), ("schedule", option_type(float64_type)), ("price", option_type(Price_reflection())), ("operator", option_type(float64_type)), ("direction", option_type(string_type)), ("line", option_type(Line_reflection())), ("reachable", option_type(bool_type)), ("cancelled", option_type(bool_type)), ("walking", option_type(bool_type)), ("load_factor", option_type(string_type)), ("distance", option_type(int32_type)), ("public", option_type(bool_type)), ("transfer", option_type(bool_type)), ("cycle", option_type(Cycle_reflection())), ("alternatives", option_type(array_type(Alternative_reflection()))), ("polyline", option_type(FeatureCollection_reflection())), ("remarks", option_type(array_type(obj_type)))])


class Trip(Record):
    def __init__(self, id: str, origin: Optional[Any], destination: Optional[Any], departure: Optional[str], planned_departure: Optional[str], prognosed_arrival: Optional[str], departure_delay: Optional[int], departure_platform: Optional[str], prognosed_departure_platform: Optional[str], planned_departure_platform: Optional[str], arrival: Optional[str], planned_arrival: Optional[str], prognosed_departure: Optional[str], arrival_delay: Optional[int], arrival_platform: Optional[str], prognosed_arrival_platform: Optional[str], planned_arrival_platform: Optional[str], stopovers: Optional[List[StopOver]], schedule: Optional[float], price: Optional[Price], operator: Optional[float], direction: Optional[str], line: Optional[Line], reachable: Optional[bool], cancelled: Optional[bool], walking: Optional[bool], load_factor: Optional[str], distance: Optional[int], public: Optional[bool], transfer: Optional[bool], cycle: Optional[Cycle], alternatives: Optional[List[Alternative]], polyline: Optional[FeatureCollection], remarks: Optional[List[Any]]) -> None:
        super().__init__()
        self.id = id
        self.origin = origin
        self.destination = destination
        self.departure = departure
        self.planned_departure = planned_departure
        self.prognosed_arrival = prognosed_arrival
        self.departure_delay = departure_delay
        self.departure_platform = departure_platform
        self.prognosed_departure_platform = prognosed_departure_platform
        self.planned_departure_platform = planned_departure_platform
        self.arrival = arrival
        self.planned_arrival = planned_arrival
        self.prognosed_departure = prognosed_departure
        self.arrival_delay = arrival_delay
        self.arrival_platform = arrival_platform
        self.prognosed_arrival_platform = prognosed_arrival_platform
        self.planned_arrival_platform = planned_arrival_platform
        self.stopovers = stopovers
        self.schedule = schedule
        self.price = price
        self.operator = operator
        self.direction = direction
        self.line = line
        self.reachable = reachable
        self.cancelled = cancelled
        self.walking = walking
        self.load_factor = load_factor
        self.distance = distance
        self.public = public
        self.transfer = transfer
        self.cycle = cycle
        self.alternatives = alternatives
        self.polyline = polyline
        self.remarks = remarks
    

Trip_reflection = expr_262

def expr_263() -> TypeInfo:
    return record_type("FsHafas.Client.Price", [], Price, lambda: [("amount", float64_type), ("currency", string_type), ("hint", option_type(string_type))])


class Price(Record):
    def __init__(self, amount: float, currency: str, hint: Optional[str]) -> None:
        super().__init__()
        self.amount = amount
        self.currency = currency
        self.hint = hint
    

Price_reflection = expr_263

def expr_264() -> TypeInfo:
    return record_type("FsHafas.Client.Alternative", [], Alternative, lambda: [("trip_id", string_type), ("direction", option_type(string_type)), ("location", option_type(Location_reflection())), ("line", option_type(Line_reflection())), ("stop", option_type(obj_type)), ("when", option_type(string_type)), ("planned_when", option_type(string_type)), ("prognosed_when", option_type(string_type)), ("delay", option_type(int32_type)), ("platform", option_type(string_type)), ("planned_platform", option_type(string_type)), ("prognosed_platform", option_type(string_type)), ("remarks", option_type(array_type(obj_type))), ("cancelled", option_type(bool_type)), ("load_factor", option_type(string_type)), ("provenance", option_type(string_type)), ("previous_stopovers", option_type(array_type(StopOver_reflection()))), ("next_stopovers", option_type(array_type(StopOver_reflection()))), ("frames", option_type(array_type(Frame_reflection()))), ("polyline", option_type(FeatureCollection_reflection())), ("current_trip_position", option_type(Location_reflection())), ("origin", option_type(obj_type)), ("destination", option_type(obj_type))])


class Alternative(Record):
    def __init__(self, trip_id: str, direction: Optional[str], location: Optional[Location], line: Optional[Line], stop: Optional[Any], when: Optional[str], planned_when: Optional[str], prognosed_when: Optional[str], delay: Optional[int], platform: Optional[str], planned_platform: Optional[str], prognosed_platform: Optional[str], remarks: Optional[List[Any]], cancelled: Optional[bool], load_factor: Optional[str], provenance: Optional[str], previous_stopovers: Optional[List[StopOver]], next_stopovers: Optional[List[StopOver]], frames: Optional[List[Frame]], polyline: Optional[FeatureCollection], current_trip_position: Optional[Location], origin: Optional[Any], destination: Optional[Any]) -> None:
        super().__init__()
        self.trip_id = trip_id
        self.direction = direction
        self.location = location
        self.line = line
        self.stop = stop
        self.when = when
        self.planned_when = planned_when
        self.prognosed_when = prognosed_when
        self.delay = delay
        self.platform = platform
        self.planned_platform = planned_platform
        self.prognosed_platform = prognosed_platform
        self.remarks = remarks
        self.cancelled = cancelled
        self.load_factor = load_factor
        self.provenance = provenance
        self.previous_stopovers = previous_stopovers
        self.next_stopovers = next_stopovers
        self.frames = frames
        self.polyline = polyline
        self.current_trip_position = current_trip_position
        self.origin = origin
        self.destination = destination
    

Alternative_reflection = expr_264

def expr_265() -> TypeInfo:
    return record_type("FsHafas.Client.Leg", [], Leg, lambda: [("trip_id", option_type(string_type)), ("origin", option_type(obj_type)), ("destination", option_type(obj_type)), ("departure", option_type(string_type)), ("planned_departure", option_type(string_type)), ("prognosed_arrival", option_type(string_type)), ("departure_delay", option_type(int32_type)), ("departure_platform", option_type(string_type)), ("prognosed_departure_platform", option_type(string_type)), ("planned_departure_platform", option_type(string_type)), ("arrival", option_type(string_type)), ("planned_arrival", option_type(string_type)), ("prognosed_departure", option_type(string_type)), ("arrival_delay", option_type(int32_type)), ("arrival_platform", option_type(string_type)), ("prognosed_arrival_platform", option_type(string_type)), ("planned_arrival_platform", option_type(string_type)), ("stopovers", option_type(array_type(StopOver_reflection()))), ("schedule", option_type(float64_type)), ("price", option_type(Price_reflection())), ("operator", option_type(float64_type)), ("direction", option_type(string_type)), ("line", option_type(Line_reflection())), ("reachable", option_type(bool_type)), ("cancelled", option_type(bool_type)), ("walking", option_type(bool_type)), ("load_factor", option_type(string_type)), ("distance", option_type(int32_type)), ("public", option_type(bool_type)), ("transfer", option_type(bool_type)), ("cycle", option_type(Cycle_reflection())), ("alternatives", option_type(array_type(Alternative_reflection()))), ("polyline", option_type(FeatureCollection_reflection())), ("remarks", option_type(array_type(obj_type))), ("current_location", option_type(Location_reflection()))])


class Leg(Record):
    def __init__(self, trip_id: Optional[str], origin: Optional[Any], destination: Optional[Any], departure: Optional[str], planned_departure: Optional[str], prognosed_arrival: Optional[str], departure_delay: Optional[int], departure_platform: Optional[str], prognosed_departure_platform: Optional[str], planned_departure_platform: Optional[str], arrival: Optional[str], planned_arrival: Optional[str], prognosed_departure: Optional[str], arrival_delay: Optional[int], arrival_platform: Optional[str], prognosed_arrival_platform: Optional[str], planned_arrival_platform: Optional[str], stopovers: Optional[List[StopOver]], schedule: Optional[float], price: Optional[Price], operator: Optional[float], direction: Optional[str], line: Optional[Line], reachable: Optional[bool], cancelled: Optional[bool], walking: Optional[bool], load_factor: Optional[str], distance: Optional[int], public: Optional[bool], transfer: Optional[bool], cycle: Optional[Cycle], alternatives: Optional[List[Alternative]], polyline: Optional[FeatureCollection], remarks: Optional[List[Any]], current_location: Optional[Location]) -> None:
        super().__init__()
        self.trip_id = trip_id
        self.origin = origin
        self.destination = destination
        self.departure = departure
        self.planned_departure = planned_departure
        self.prognosed_arrival = prognosed_arrival
        self.departure_delay = departure_delay
        self.departure_platform = departure_platform
        self.prognosed_departure_platform = prognosed_departure_platform
        self.planned_departure_platform = planned_departure_platform
        self.arrival = arrival
        self.planned_arrival = planned_arrival
        self.prognosed_departure = prognosed_departure
        self.arrival_delay = arrival_delay
        self.arrival_platform = arrival_platform
        self.prognosed_arrival_platform = prognosed_arrival_platform
        self.planned_arrival_platform = planned_arrival_platform
        self.stopovers = stopovers
        self.schedule = schedule
        self.price = price
        self.operator = operator
        self.direction = direction
        self.line = line
        self.reachable = reachable
        self.cancelled = cancelled
        self.walking = walking
        self.load_factor = load_factor
        self.distance = distance
        self.public = public
        self.transfer = transfer
        self.cycle = cycle
        self.alternatives = alternatives
        self.polyline = polyline
        self.remarks = remarks
        self.current_location = current_location
    

Leg_reflection = expr_265

def expr_266() -> TypeInfo:
    return record_type("FsHafas.Client.Journey", [], Journey, lambda: [("type", option_type(string_type)), ("legs", array_type(Leg_reflection())), ("refresh_token", option_type(string_type)), ("remarks", option_type(array_type(obj_type))), ("price", option_type(Price_reflection())), ("cycle", option_type(Cycle_reflection())), ("scheduled_days", option_type(IndexMap_2_reflection(string_type, bool_type)))])


class Journey(Record):
    def __init__(self, type: Optional[str], legs: List[Leg], refresh_token: Optional[str], remarks: Optional[List[Any]], price: Optional[Price], cycle: Optional[Cycle], scheduled_days: Optional[IndexMap_2[str, bool]]) -> None:
        super().__init__()
        self.type = type
        self.legs = legs
        self.refresh_token = refresh_token
        self.remarks = remarks
        self.price = price
        self.cycle = cycle
        self.scheduled_days = scheduled_days
    

Journey_reflection = expr_266

def expr_267() -> TypeInfo:
    return record_type("FsHafas.Client.Journeys", [], Journeys, lambda: [("earlier_ref", option_type(string_type)), ("later_ref", option_type(string_type)), ("journeys", option_type(array_type(Journey_reflection()))), ("realtime_data_from", option_type(int32_type))])


class Journeys(Record):
    def __init__(self, earlier_ref: Optional[str], later_ref: Optional[str], journeys: Optional[List[Journey]], realtime_data_from: Optional[int]) -> None:
        super().__init__()
        self.earlier_ref = earlier_ref
        self.later_ref = later_ref
        self.journeys = journeys
        self.realtime_data_from = realtime_data_from
    

Journeys_reflection = expr_267

def expr_268() -> TypeInfo:
    return record_type("FsHafas.Client.Duration", [], Duration, lambda: [("duration", int32_type), ("stations", array_type(obj_type))])


class Duration(Record):
    def __init__(self, duration: int, stations: List[Any]) -> None:
        super().__init__()
        self.duration = duration or 0
        self.stations = stations
    

Duration_reflection = expr_268

def expr_269() -> TypeInfo:
    return record_type("FsHafas.Client.Frame", [], Frame, lambda: [("origin", obj_type), ("destination", obj_type), ("t", option_type(int32_type))])


class Frame(Record):
    def __init__(self, origin: Any, destination: Any, t: Optional[int]) -> None:
        super().__init__()
        self.origin = origin
        self.destination = destination
        self.t = t
    

Frame_reflection = expr_269

def expr_270() -> TypeInfo:
    return record_type("FsHafas.Client.Movement", [], Movement, lambda: [("direction", option_type(string_type)), ("trip_id", option_type(string_type)), ("line", option_type(Line_reflection())), ("location", option_type(Location_reflection())), ("next_stopovers", option_type(array_type(StopOver_reflection()))), ("frames", option_type(array_type(Frame_reflection()))), ("polyline", option_type(FeatureCollection_reflection()))])


class Movement(Record):
    def __init__(self, direction: Optional[str], trip_id: Optional[str], line: Optional[Line], location: Optional[Location], next_stopovers: Optional[List[StopOver]], frames: Optional[List[Frame]], polyline: Optional[FeatureCollection]) -> None:
        super().__init__()
        self.direction = direction
        self.trip_id = trip_id
        self.line = line
        self.location = location
        self.next_stopovers = next_stopovers
        self.frames = frames
        self.polyline = polyline
    

Movement_reflection = expr_270

def expr_271() -> TypeInfo:
    return record_type("FsHafas.Client.ServerInfo", [], ServerInfo, lambda: [("hci_version", option_type(string_type)), ("timetable_start", option_type(string_type)), ("timetable_end", option_type(string_type)), ("server_time", option_type(string_type)), ("realtime_data_updated_at", option_type(int32_type))])


class ServerInfo(Record):
    def __init__(self, hci_version: Optional[str], timetable_start: Optional[str], timetable_end: Optional[str], server_time: Optional[str], realtime_data_updated_at: Optional[int]) -> None:
        super().__init__()
        self.hci_version = hci_version
        self.timetable_start = timetable_start
        self.timetable_end = timetable_end
        self.server_time = server_time
        self.realtime_data_updated_at = realtime_data_updated_at
    

ServerInfo_reflection = expr_271

def expr_272() -> TypeInfo:
    return record_type("FsHafas.Client.LoyaltyCard", [], LoyaltyCard, lambda: [("type", option_type(string_type)), ("discount", option_type(int32_type)), ("class_", option_type(int32_type))])


class LoyaltyCard(Record):
    def __init__(self, type: Optional[str], discount: Optional[int], class_: Optional[int]) -> None:
        super().__init__()
        self.type = type
        self.discount = discount
        self.class_ = class_
    

LoyaltyCard_reflection = expr_272

def expr_273() -> TypeInfo:
    return record_type("FsHafas.Client.JourneysOptions", [], JourneysOptions, lambda: [("departure", option_type(class_type("System.DateTime"))), ("arrival", option_type(class_type("System.DateTime"))), ("earlier_than", option_type(string_type)), ("later_than", option_type(string_type)), ("results", option_type(int32_type)), ("via", option_type(string_type)), ("stopovers", option_type(bool_type)), ("transfers", option_type(int32_type)), ("transfer_time", option_type(int32_type)), ("accessibility", option_type(string_type)), ("bike", option_type(bool_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("tickets", option_type(bool_type)), ("polylines", option_type(bool_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("remarks", option_type(bool_type)), ("walking_speed", option_type(string_type)), ("start_with_walking", option_type(bool_type)), ("language", option_type(string_type)), ("scheduled_days", option_type(bool_type)), ("first_class", option_type(bool_type)), ("age", option_type(int32_type)), ("loyalty_card", option_type(LoyaltyCard_reflection())), ("when", option_type(class_type("System.DateTime")))])


class JourneysOptions(Record):
    def __init__(self, departure: Optional[Any], arrival: Optional[Any], earlier_than: Optional[str], later_than: Optional[str], results: Optional[int], via: Optional[str], stopovers: Optional[bool], transfers: Optional[int], transfer_time: Optional[int], accessibility: Optional[str], bike: Optional[bool], products: Optional[IndexMap_2[str, bool]], tickets: Optional[bool], polylines: Optional[bool], sub_stops: Optional[bool], entrances: Optional[bool], remarks: Optional[bool], walking_speed: Optional[str], start_with_walking: Optional[bool], language: Optional[str], scheduled_days: Optional[bool], first_class: Optional[bool], age: Optional[int], loyalty_card: Optional[LoyaltyCard], when: Optional[Any]) -> None:
        super().__init__()
        self.departure = departure
        self.arrival = arrival
        self.earlier_than = earlier_than
        self.later_than = later_than
        self.results = results
        self.via = via
        self.stopovers = stopovers
        self.transfers = transfers
        self.transfer_time = transfer_time
        self.accessibility = accessibility
        self.bike = bike
        self.products = products
        self.tickets = tickets
        self.polylines = polylines
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.remarks = remarks
        self.walking_speed = walking_speed
        self.start_with_walking = start_with_walking
        self.language = language
        self.scheduled_days = scheduled_days
        self.first_class = first_class
        self.age = age
        self.loyalty_card = loyalty_card
        self.when = when
    

JourneysOptions_reflection = expr_273

def expr_274() -> TypeInfo:
    return record_type("FsHafas.Client.JourneysFromTripOptions", [], JourneysFromTripOptions, lambda: [("stopovers", option_type(bool_type)), ("transfer_time", option_type(int32_type)), ("accessibility", option_type(string_type)), ("tickets", option_type(bool_type)), ("polylines", option_type(bool_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("remarks", option_type(bool_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type)))])


class JourneysFromTripOptions(Record):
    def __init__(self, stopovers: Optional[bool], transfer_time: Optional[int], accessibility: Optional[str], tickets: Optional[bool], polylines: Optional[bool], sub_stops: Optional[bool], entrances: Optional[bool], remarks: Optional[bool], products: Optional[IndexMap_2[str, bool]]) -> None:
        super().__init__()
        self.stopovers = stopovers
        self.transfer_time = transfer_time
        self.accessibility = accessibility
        self.tickets = tickets
        self.polylines = polylines
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.remarks = remarks
        self.products = products
    

JourneysFromTripOptions_reflection = expr_274

def expr_275() -> TypeInfo:
    return record_type("FsHafas.Client.LocationsOptions", [], LocationsOptions, lambda: [("fuzzy", option_type(bool_type)), ("results", option_type(int32_type)), ("stops", option_type(bool_type)), ("addresses", option_type(bool_type)), ("poi", option_type(bool_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("lines_of_stops", option_type(bool_type)), ("language", option_type(string_type))])


class LocationsOptions(Record):
    def __init__(self, fuzzy: Optional[bool], results: Optional[int], stops: Optional[bool], addresses: Optional[bool], poi: Optional[bool], sub_stops: Optional[bool], entrances: Optional[bool], lines_of_stops: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.fuzzy = fuzzy
        self.results = results
        self.stops = stops
        self.addresses = addresses
        self.poi = poi
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.lines_of_stops = lines_of_stops
        self.language = language
    

LocationsOptions_reflection = expr_275

def expr_276() -> TypeInfo:
    return record_type("FsHafas.Client.TripOptions", [], TripOptions, lambda: [("stopovers", option_type(bool_type)), ("polyline", option_type(bool_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("remarks", option_type(bool_type)), ("language", option_type(string_type))])


class TripOptions(Record):
    def __init__(self, stopovers: Optional[bool], polyline: Optional[bool], sub_stops: Optional[bool], entrances: Optional[bool], remarks: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.stopovers = stopovers
        self.polyline = polyline
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.remarks = remarks
        self.language = language
    

TripOptions_reflection = expr_276

def expr_277() -> TypeInfo:
    return record_type("FsHafas.Client.StopOptions", [], StopOptions, lambda: [("lines_of_stops", option_type(bool_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("remarks", option_type(bool_type)), ("language", option_type(string_type))])


class StopOptions(Record):
    def __init__(self, lines_of_stops: Optional[bool], sub_stops: Optional[bool], entrances: Optional[bool], remarks: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.lines_of_stops = lines_of_stops
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.remarks = remarks
        self.language = language
    

StopOptions_reflection = expr_277

def expr_278() -> TypeInfo:
    return record_type("FsHafas.Client.DeparturesArrivalsOptions", [], DeparturesArrivalsOptions, lambda: [("when", option_type(class_type("System.DateTime"))), ("direction", option_type(string_type)), ("line", option_type(string_type)), ("duration", option_type(int32_type)), ("results", option_type(int32_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("lines_of_stops", option_type(bool_type)), ("remarks", option_type(bool_type)), ("stopovers", option_type(bool_type)), ("include_related_stations", option_type(bool_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("language", option_type(string_type))])


class DeparturesArrivalsOptions(Record):
    def __init__(self, when: Optional[Any], direction: Optional[str], line: Optional[str], duration: Optional[int], results: Optional[int], sub_stops: Optional[bool], entrances: Optional[bool], lines_of_stops: Optional[bool], remarks: Optional[bool], stopovers: Optional[bool], include_related_stations: Optional[bool], products: Optional[IndexMap_2[str, bool]], language: Optional[str]) -> None:
        super().__init__()
        self.when = when
        self.direction = direction
        self.line = line
        self.duration = duration
        self.results = results
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.lines_of_stops = lines_of_stops
        self.remarks = remarks
        self.stopovers = stopovers
        self.include_related_stations = include_related_stations
        self.products = products
        self.language = language
    

DeparturesArrivalsOptions_reflection = expr_278

def expr_279() -> TypeInfo:
    return record_type("FsHafas.Client.RefreshJourneyOptions", [], RefreshJourneyOptions, lambda: [("stopovers", option_type(bool_type)), ("polylines", option_type(bool_type)), ("tickets", option_type(bool_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("remarks", option_type(bool_type)), ("language", option_type(string_type))])


class RefreshJourneyOptions(Record):
    def __init__(self, stopovers: Optional[bool], polylines: Optional[bool], tickets: Optional[bool], sub_stops: Optional[bool], entrances: Optional[bool], remarks: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.stopovers = stopovers
        self.polylines = polylines
        self.tickets = tickets
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.remarks = remarks
        self.language = language
    

RefreshJourneyOptions_reflection = expr_279

def expr_280() -> TypeInfo:
    return record_type("FsHafas.Client.NearByOptions", [], NearByOptions, lambda: [("results", option_type(int32_type)), ("distance", option_type(int32_type)), ("poi", option_type(bool_type)), ("stops", option_type(bool_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("lines_of_stops", option_type(bool_type)), ("language", option_type(string_type))])


class NearByOptions(Record):
    def __init__(self, results: Optional[int], distance: Optional[int], poi: Optional[bool], stops: Optional[bool], products: Optional[IndexMap_2[str, bool]], sub_stops: Optional[bool], entrances: Optional[bool], lines_of_stops: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.results = results
        self.distance = distance
        self.poi = poi
        self.stops = stops
        self.products = products
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.lines_of_stops = lines_of_stops
        self.language = language
    

NearByOptions_reflection = expr_280

def expr_281() -> TypeInfo:
    return record_type("FsHafas.Client.ReachableFromOptions", [], ReachableFromOptions, lambda: [("when", option_type(class_type("System.DateTime"))), ("max_transfers", option_type(int32_type)), ("max_duration", option_type(int32_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("polylines", option_type(bool_type))])


class ReachableFromOptions(Record):
    def __init__(self, when: Optional[Any], max_transfers: Optional[int], max_duration: Optional[int], products: Optional[IndexMap_2[str, bool]], sub_stops: Optional[bool], entrances: Optional[bool], polylines: Optional[bool]) -> None:
        super().__init__()
        self.when = when
        self.max_transfers = max_transfers
        self.max_duration = max_duration
        self.products = products
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.polylines = polylines
    

ReachableFromOptions_reflection = expr_281

def expr_282() -> TypeInfo:
    return record_type("FsHafas.Client.BoundingBox", [], BoundingBox, lambda: [("north", float64_type), ("west", float64_type), ("south", float64_type), ("east", float64_type)])


class BoundingBox(Record):
    def __init__(self, north: float, west: float, south: float, east: float) -> None:
        super().__init__()
        self.north = north
        self.west = west
        self.south = south
        self.east = east
    

BoundingBox_reflection = expr_282

def expr_283() -> TypeInfo:
    return record_type("FsHafas.Client.RadarOptions", [], RadarOptions, lambda: [("results", option_type(int32_type)), ("frames", option_type(int32_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("duration", option_type(int32_type)), ("sub_stops", option_type(bool_type)), ("entrances", option_type(bool_type)), ("polylines", option_type(bool_type)), ("when", option_type(class_type("System.DateTime")))])


class RadarOptions(Record):
    def __init__(self, results: Optional[int], frames: Optional[int], products: Optional[IndexMap_2[str, bool]], duration: Optional[int], sub_stops: Optional[bool], entrances: Optional[bool], polylines: Optional[bool], when: Optional[Any]) -> None:
        super().__init__()
        self.results = results
        self.frames = frames
        self.products = products
        self.duration = duration
        self.sub_stops = sub_stops
        self.entrances = entrances
        self.polylines = polylines
        self.when = when
    

RadarOptions_reflection = expr_283

def expr_284() -> TypeInfo:
    return record_type("FsHafas.Client.Filter", [], Filter, lambda: [("type", option_type(string_type)), ("mode", string_type), ("value", string_type)])


class Filter(Record):
    def __init__(self, type: Optional[str], mode: str, value: str) -> None:
        super().__init__()
        self.type = type
        self.mode = mode
        self.value = value
    

Filter_reflection = expr_284

def expr_285() -> TypeInfo:
    return record_type("FsHafas.Client.TripsByNameOptions", [], TripsByNameOptions, lambda: [("when", option_type(class_type("System.DateTime"))), ("from_when", option_type(class_type("System.DateTime"))), ("until_when", option_type(class_type("System.DateTime"))), ("only_currently_running", option_type(bool_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("currently_stopping_at", option_type(obj_type)), ("line_name", option_type(string_type)), ("operator_names", option_type(array_type(string_type))), ("additional_filters", option_type(array_type(Filter_reflection())))])


class TripsByNameOptions(Record):
    def __init__(self, when: Optional[Any], from_when: Optional[Any], until_when: Optional[Any], only_currently_running: Optional[bool], products: Optional[IndexMap_2[str, bool]], currently_stopping_at: Optional[Any], line_name: Optional[str], operator_names: Optional[List[str]], additional_filters: Optional[List[Filter]]) -> None:
        super().__init__()
        self.when = when
        self.from_when = from_when
        self.until_when = until_when
        self.only_currently_running = only_currently_running
        self.products = products
        self.currently_stopping_at = currently_stopping_at
        self.line_name = line_name
        self.operator_names = operator_names
        self.additional_filters = additional_filters
    

TripsByNameOptions_reflection = expr_285

def expr_286() -> TypeInfo:
    return record_type("FsHafas.Client.RemarksOptions", [], RemarksOptions, lambda: [("from_", option_type(class_type("System.DateTime"))), ("to", option_type(class_type("System.DateTime"))), ("results", option_type(int32_type)), ("products", option_type(IndexMap_2_reflection(string_type, bool_type))), ("polylines", option_type(bool_type)), ("language", option_type(string_type))])


class RemarksOptions(Record):
    def __init__(self, from_: Optional[Any], to: Optional[Any], results: Optional[int], products: Optional[IndexMap_2[str, bool]], polylines: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.from_ = from_
        self.to = to
        self.results = results
        self.products = products
        self.polylines = polylines
        self.language = language
    

RemarksOptions_reflection = expr_286

def expr_287() -> TypeInfo:
    return record_type("FsHafas.Client.LinesOptions", [], LinesOptions, lambda: [("language", option_type(string_type))])


class LinesOptions(Record):
    def __init__(self, language: Optional[str]=None) -> None:
        super().__init__()
        self.language = language
    

LinesOptions_reflection = expr_287

def expr_288() -> TypeInfo:
    return record_type("FsHafas.Client.ServerOptions", [], ServerOptions, lambda: [("version_info", option_type(bool_type)), ("language", option_type(string_type))])


class ServerOptions(Record):
    def __init__(self, version_info: Optional[bool], language: Optional[str]) -> None:
        super().__init__()
        self.version_info = version_info
        self.language = language
    

ServerOptions_reflection = expr_288

class HafasClient(Protocol):
    @abstractmethod
    def arrivals(self, __arg0: Any, __arg1: Optional[DeparturesArrivalsOptions]) -> Promise_1[List[Alternative]]:
        ...
    
    @abstractmethod
    def departures(self, __arg0: Any, __arg1: Optional[DeparturesArrivalsOptions]) -> Promise_1[List[Alternative]]:
        ...
    
    @abstractmethod
    def journeys(self, __arg0: Any, __arg1: Any, __arg2: Optional[JourneysOptions]) -> Promise_1[Journeys]:
        ...
    
    @abstractmethod
    def journeys_from_trip(self, __arg0: str, __arg1: StopOver, __arg2: Any, __arg3: Optional[JourneysFromTripOptions]) -> Promise_1[List[Journey]]:
        ...
    
    @abstractmethod
    def lines(self, __arg0: str, __arg1: Optional[LinesOptions]) -> Promise_1[List[Line]]:
        ...
    
    @abstractmethod
    def locations(self, __arg0: str, __arg1: Optional[LocationsOptions]) -> Promise_1[List[Any]]:
        ...
    
    @abstractmethod
    def nearby(self, __arg0: Location, __arg1: Optional[NearByOptions]) -> Promise_1[List[Any]]:
        ...
    
    @abstractmethod
    def radar(self, __arg0: BoundingBox, __arg1: Optional[RadarOptions]) -> Promise_1[List[Movement]]:
        ...
    
    @abstractmethod
    def reachable_from(self, __arg0: Location, __arg1: Optional[ReachableFromOptions]) -> Promise_1[List[Duration]]:
        ...
    
    @abstractmethod
    def refresh_journey(self, __arg0: str, __arg1: Optional[RefreshJourneyOptions]) -> Promise_1[Journey]:
        ...
    
    @abstractmethod
    def remarks(self, __arg0: Optional[RemarksOptions]) -> Promise_1[List[Warning]]:
        ...
    
    @abstractmethod
    def server_info(self, __arg0: Optional[ServerOptions]) -> Promise_1[ServerInfo]:
        ...
    
    @abstractmethod
    def stop(self, __arg0: Any, __arg1: Optional[StopOptions]) -> Promise_1[Any]:
        ...
    
    @abstractmethod
    def trip(self, __arg0: str, __arg1: str, __arg2: Optional[TripOptions]) -> Promise_1[Trip]:
        ...
    
    @abstractmethod
    def trips_by_name(self, __arg0: str, __arg1: Optional[TripsByNameOptions]) -> Promise_1[List[Trip]]:
        ...
    




from datetime import datetime, timedelta
from typing import Union

from .types import FSharpRef


def add(d: datetime, ts: timedelta) -> datetime:
    return d + ts


def parse(string: str, detectUTC: bool = False) -> datetime:
    from dateutil import parser

    return parser.parse(string)


def try_parse(
    string: str, style: int, unsigned: bool, bitsize: int, defValue: FSharpRef[datetime]
) -> bool:
    try:
        defValue.contents = parse(string)
        return True
    except Exception:
        return False


def now() -> datetime:
    return datetime.now()


def utc_now() -> datetime:
    return datetime.utcnow()


def op_addition(x: datetime, y: timedelta) -> datetime:
    return x + y


def op_subtraction(
    x: datetime, y: Union[datetime, timedelta]
) -> Union[datetime, timedelta]:
    if isinstance(y, timedelta):
        return x - y

    return x - y


__all__ = ["now", "op_addition", "op_subtraction", "parse", "try_parse", "utc_now"]



from typing import (TypeVar, Callable, Any, Optional)
from .util import (IEqualityComparer, structural_hash, equals, physical_hash, IComparer, compare)

_T = TypeVar("_T")

_T_ = TypeVar("_T_")

def HashIdentity_FromFunctions(hash_1: Callable[[_T], int], eq: Callable[[_T, _T], bool]) -> IEqualityComparer[Any]:
    class ObjectExpr1(IEqualityComparer[Any]):
        def Equals(self, x: _T_, y: _T_, hash_1: Callable[[_T], int]=hash_1, eq: Callable[[_T, _T], bool]=eq) -> bool:
            return eq(x, y)
        
        def GetHashCode(self, x_1: Optional[_T_]=None, hash_1: Callable[[_T], int]=hash_1, eq: Callable[[_T, _T], bool]=eq) -> int:
            return hash_1(x_1)
        
    return ObjectExpr1()


def HashIdentity_Structural() -> IEqualityComparer[Any]:
    def arrow_2(obj: Optional[_T]=None) -> int:
        return structural_hash(obj)
    
    def arrow_3(e1: _T, e2: _T) -> bool:
        return equals(e1, e2)
    
    return HashIdentity_FromFunctions(arrow_2, arrow_3)


def HashIdentity_Reference() -> IEqualityComparer[Any]:
    def arrow_4(obj: Optional[_T]=None) -> int:
        return physical_hash(obj)
    
    def arrow_5(e1: _T, e2: _T) -> bool:
        return e1 is e2
    
    return HashIdentity_FromFunctions(arrow_4, arrow_5)


def ComparisonIdentity_FromFunction(comparer: Callable[[_T, _T], int]) -> IComparer[_T]:
    class ObjectExpr6(IComparer[_T_]):
        def Compare(self, x: _T_, y: _T_, comparer: Callable[[_T, _T], int]=comparer) -> int:
            return comparer(x, y)
        
    return ObjectExpr6()


def ComparisonIdentity_Structural() -> IComparer[_T]:
    def arrow_7(e1: _T, e2: _T) -> int:
        return compare(e1, e2)
    
    return ComparisonIdentity_FromFunction(arrow_7)





from __future__ import annotations
from typing import (Any, Optional, Callable, Tuple, List)
from ...fable_library.array import (map, filter)
from ...fable_library.option import default_arg
from ...fs_hafas_python.context import (Context, Profile__get_parseWhen, ParsedWhen, Profile__get_parsePlatform, Platform, Profile__get_parseStopover)
from ...fs_hafas_python.lib.transformations import U2StationStop_FromSomeU3StationStopLocation
from ...fs_hafas_python.types_hafas_client import StopOver
from ...fs_hafas_python.types_raw_hafas_client import (RawStop, RawPltf)
from .common import (get_element_at, msg_lto_remarks)

default_stopover : StopOver = StopOver(None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None, None)

def parse_stopover(ctx: Context, st: RawStop, date: str) -> StopOver:
    stop : Optional[Any] = U2StationStop_FromSomeU3StationStopLocation(get_element_at(st.loc_x, ctx.common.locations))
    dep : ParsedWhen = Profile__get_parseWhen(ctx.profile)(ctx)(date)(st.d_time_s)(st.d_time_r)(st.d_tzoffset)(st.d_cncl)
    arr : ParsedWhen = Profile__get_parseWhen(ctx.profile)(ctx)(date)(st.a_time_s)(st.a_time_r)(st.a_tzoffset)(st.a_cncl)
    def match_platf_s(a_platf_s: Optional[str]=None, a_pltf_s: Optional[RawPltf]=None, ctx: Context=ctx, st: RawStop=st, date: str=date) -> Optional[str]:
        if a_platf_s is None:
            if a_pltf_s is not None:
                return a_pltf_s.txt
            
            else: 
                return None
            
        
        else: 
            return a_platf_s
        
    
    d_platf_s : Optional[str] = match_platf_s(st.d_platf_s, st.d_pltf_s)
    d_platf_r : Optional[str] = match_platf_s(st.d_platf_r, st.d_pltf_r)
    dep_pl : Platform = Profile__get_parsePlatform(ctx.profile)(ctx)(d_platf_s)(d_platf_r)(st.d_cncl)
    a_platf_s_1 : Optional[str] = match_platf_s(st.a_platf_s, st.a_pltf_s)
    a_platf_r : Optional[str] = match_platf_s(st.a_platf_r, st.a_pltf_r)
    arr_pl : Platform = Profile__get_parsePlatform(ctx.profile)(ctx)(a_platf_s_1)(a_platf_r)(st.a_cncl)
    pass_by : Optional[bool]
    match_value : Tuple[Optional[bool], Optional[bool]] = (st.d_in_s, st.a_out_s)
    (pattern_matching_result,) = (None,)
    if match_value[0] is not None:
        if match_value[0]:
            pattern_matching_result = 1
        
        elif match_value[1] is not None:
            if match_value[1]:
                pattern_matching_result = 1
            
            else: 
                pattern_matching_result = 0
            
        
        else: 
            pattern_matching_result = 1
        
    
    else: 
        pattern_matching_result = 1
    
    if pattern_matching_result == 0:
        pass_by = True
    
    elif pattern_matching_result == 1:
        pass_by = None
    
    cancelled : Optional[bool] = default_arg(st.a_cncl, st.d_cncl)
    return StopOver(stop, dep.when, dep.delay, default_stopover.prognosed_departure, dep.planned_when, dep_pl.platform, dep_pl.prognosed_platform, dep_pl.planned_platform, arr.when, arr.delay, default_stopover.prognosed_arrival, arr.planned_when, arr_pl.platform, arr_pl.prognosed_platform, arr_pl.planned_platform, msg_lto_remarks(ctx, st.msg_l), pass_by, cancelled)


def parse_stopovers(ctx: Context, stop_l: Optional[List[RawStop]], date: str) -> Optional[List[StopOver]]:
    if stop_l is None:
        return None
    
    else: 
        def mapping(s_1: RawStop, ctx: Context=ctx, stop_l: Optional[List[RawStop]]=stop_l, date: str=date) -> StopOver:
            return Profile__get_parseStopover(ctx.profile)(ctx)(s_1)(date)
        
        def predicate(s: RawStop, ctx: Context=ctx, stop_l: Optional[List[RawStop]]=stop_l, date: str=date) -> bool:
            return get_element_at(s.loc_x, ctx.common.locations) is not None
        
        return map(mapping, filter(predicate, stop_l), None)
    





import locale
import re
from abc import ABC
from base64 import b64decode, b64encode
from dataclasses import dataclass
from datetime import datetime
from enum import Enum
from typing import (
    Any,
    Callable,
    Iterable,
    List,
    Match,
    NoReturn,
    Optional,
    Pattern,
    TypeVar,
    Union,
    overload,
)

from .date import to_string as date_to_string
from .numeric import multiply, to_exponential, to_fixed, to_hex, to_precision
from .reg_exp import escape
from .types import to_string

T = TypeVar("T")


_fs_format_regexp: Pattern[str] = re.compile(
    r"(^|[^%])%([0+\- ]*)(\d+)?(?:\.(\d+))?(\w)"
)
_interpolate_regexp: Pattern[str] = re.compile(
    r"(?:(^|[^%])%([0+\- ]*)(\d+)?(?:\.(\d+))?(\w))?%P\(\)"
)
_format_regexp: Pattern[str] = re.compile(
    r"\{(\d+)(,-?\d+)?(?:\:([a-zA-Z])(\d{0,2})|\:(.+?))?\}"
)


IPrintfFormatContinuation = Callable[[Callable[[str], Any]], Callable[[str], Any]]


@dataclass
class IPrintfFormat(ABC):
    input: str
    cont: IPrintfFormatContinuation


def printf(input: str) -> IPrintfFormat:
    # print("printf: ", input)
    format: IPrintfFormatContinuation = fs_format(input)
    return IPrintfFormat(input=input, cont=format)


def continue_print(cont: Callable[[str], Any], arg: Union[IPrintfFormat, str]) -> Any:
    """Print continuation."""

    if isinstance(arg, IPrintfFormat):
        ret = arg.cont(cont)
        return ret

    return cont(arg)


def to_console(arg: Union[IPrintfFormat, Any]) -> Any:
    return continue_print(print, arg)


def to_console_error(arg: Union[IPrintfFormat, str]):
    return continue_print(lambda x: print(x), arg)


# Set return to `Any` since `Union[str, Callable]` will make type
# checkers really unhappy.
def to_text(arg: Union[IPrintfFormat, str]) -> Any:
    cont: Callable[[str], Any] = lambda x: x
    return continue_print(cont, arg)


def to_fail(arg: Union[IPrintfFormat, str]):
    def fail(msg: str):
        raise Exception(msg)

    return continue_print(fail, arg)


def format_replacement(
    rep: Any, flags: Any, padLength: Any, precision: Any, format: Any
):
    sign = ""
    flags = flags or ""
    format = format or ""

    if isinstance(rep, (int, float)):
        if format.lower() != "x":
            if rep < 0:
                rep = rep * -1
                sign = "-"
            else:
                if flags.find(" ") >= 0:
                    sign = " "
                elif flags.find("+") >= 0:
                    sign = "+"

        elif format == "x":
            rep = to_hex(rep)
        elif format == "X":
            rep = to_hex(rep).upper()

        precision = None if precision is None else int(precision)
        if format in ("f", "F"):
            precision = precision if precision is not None else 6
            rep = to_fixed(rep, precision)
        elif format in ("g", "G"):
            rep = (
                to_precision(rep, precision)
                if precision is not None
                else to_precision(rep)
            )
        elif format in ("e", "E"):
            rep = (
                to_exponential(rep, precision)
                if precision is not None
                else to_exponential(rep)
            )
        else:  # AOid
            rep = to_string(rep)

    elif isinstance(rep, datetime):
        rep = date_to_string(rep)
    else:
        rep = to_string(rep)

    if padLength is not None:
        padLength = int(padLength)
        zeroFlag = flags.find("0") >= 0  # Use '0' for left padding
        minusFlag = flags.find("-") >= 0  # Right padding
        ch = " " if minusFlag or not zeroFlag else "0"
        if ch == "0":
            rep = pad_left(rep, padLength - len(sign), ch, minusFlag)
            rep = sign + rep
        else:
            rep = pad_left(sign + rep, padLength, ch, minusFlag)

    else:
        rep = sign + rep

    return rep


def interpolate(string: str, values: List[Any]) -> str:
    valIdx = 0
    strIdx = 0
    result = ""
    matches = _interpolate_regexp.finditer(string)
    for match in matches:
        # The first group corresponds to the no-escape char (^|[^%]), the actual pattern starts in the next char
        # Note: we don't use negative lookbehind because some browsers don't support it yet
        matchIndex = match.start() + len(match[1] or "")
        result += string[strIdx:matchIndex].replace("%%", "%")
        [_, flags, padLength, precision, format] = match.groups()
        # print(match.groups())
        result += format_replacement(
            values[valIdx], flags, padLength, precision, format
        )
        valIdx += 1

        strIdx = match.end()

    result += string[strIdx:].replace("%%", "%")
    return result


def format_once(str2: str, rep: Any):
    def match(m: Match[str]):
        prefix, flags, padLength, precision, format = m.groups()
        once: str = format_replacement(rep, flags, padLength, precision, format)
        return prefix + once.replace("%", "%%")

    ret = _fs_format_regexp.sub(match, str2, count=1)
    return ret


def create_printer(string: str, cont: Callable[..., Any]):
    def _(*args: Any):
        strCopy: str = string
        for arg in args:
            strCopy = format_once(strCopy, arg)

        if _fs_format_regexp.search(strCopy):
            return create_printer(strCopy, cont)
        return cont(strCopy.replace("%%", "%"))

    return _


def fs_format(str: str):
    def _(cont: Callable[..., Any]):
        if _fs_format_regexp.search(str):
            return create_printer(str, cont)
        return cont(str)

    return _


def format(string: str, *args: Any) -> str:
    def match(m: Match[str]) -> str:
        idx, padLength, format, precision_, pattern = list(m.groups())
        rep = args[int(idx)]
        if isinstance(rep, (int, float)):
            precision: Optional[int] = None
            try:
                precision: Optional[int] = int(precision_)
            except Exception:
                pass

            if format in ["f", "F"]:
                precision = precision if precision is not None else 2
                rep = to_fixed(rep, precision)

            elif format in ["g", "G"]:
                rep = (
                    to_precision(rep, precision)
                    if precision is not None
                    else to_precision(rep)
                )

            elif format in ["e", "E"]:
                rep = (
                    to_exponential(rep, precision)
                    if precision is not None
                    else to_exponential(rep)
                )

            elif format in ["p", "P"]:
                precision = precision if precision is not None else 2
                rep = to_fixed(multiply(rep, 100), precision) + " %"

            elif format in ["d", "D"]:
                rep = (
                    pad_left(str(rep), precision, "0")
                    if precision is not None
                    else str(rep)
                )

            elif format in ["x", "X"]:
                rep = (
                    pad_left(to_hex(rep), precision, "0")
                    if precision is not None
                    else to_hex(rep)
                )
                if format == "X":
                    rep = rep.upper()
            elif pattern:
                sign = ""

                def match(m: Match[str]):
                    nonlocal sign, rep

                    intPart, decimalPart = list(m.groups())
                    if rep < 0:
                        rep = multiply(rep, -1)
                        sign = "-"

                    rep = to_fixed(rep, len(decimalPart) - 1 if decimalPart else 0)
                    return pad_left(
                        rep,
                        len(intPart or "")
                        - len(sign)
                        + (len(decimalPart) if decimalPart else 0),
                        "0",
                    )

                rep = re.sub(r"(0+)(\.0+)?", match, pattern)
                rep = sign + rep

        elif isinstance(rep, datetime):
            rep = date_to_string(rep, pattern or format)
        else:
            rep = to_string(rep)

        try:
            padLength = int((padLength or " ")[1:])
            rep = pad_left(str(rep), abs(padLength), " ", padLength < 0)
        except ValueError:
            pass

        return str(rep)

    return _format_regexp.sub(match, string)


def initialize(n: int, f: Callable[[int], str]) -> str:
    if n < 0:
        raise Exception("String length must be non-negative")

    xs = [""] * n
    for i in range(n):
        xs[i] = f(i)

    return "".join(xs)


def insert(string: str, startIndex: int, value: str):
    if startIndex < 0 or startIndex > len(string):
        raise ValueError(
            "startIndex is negative or greater than the length of this instance."
        )

    return string[:startIndex] + value + string[startIndex:]


def is_null_or_empty(string: Optional[str]):
    return not isinstance(string, str) or not len(string)


def is_null_or_white_space(string: Optional[Any]) -> bool:
    return not string or not isinstance(string, str) or string.isspace()


def concat(*xs: Iterable[Any]) -> str:
    return "".join(map(str, xs))


def join(delimiter: str, xs: Iterable[Any]) -> str:
    return delimiter.join((str(x) for x in xs))


def join_with_indices(delimiter: str, xs: List[str], startIndex: int, count: int):
    endIndexPlusOne = startIndex + count
    if endIndexPlusOne > len(xs):
        raise ValueError("Index and count must refer to a location within the buffer.")

    return delimiter.join(xs[startIndex:endIndexPlusOne])


def not_supported(name: str) -> NoReturn:
    raise Exception(
        "The environment doesn't support '" + name + "', please use a polyfill."
    )


def to_base64string(in_array: bytes) -> str:
    return b64encode(in_array).decode("utf8")


def from_base64string(b64encoded: str) -> bytes:
    return b64decode(b64encoded)


def pad_left(
    string: str, length: int, ch: Optional[str] = None, isRight: Optional[bool] = False
) -> str:
    ch = ch or " "
    length = length - len(string)
    for _ in range(length):
        string = string + ch if isRight else ch + string

    return string


def pad_right(string: str, len: int, ch: Optional[str] = None) -> str:
    return pad_left(string, len, ch, True)


def remove(string: str, startIndex: int, count: Optional[int] = None):
    if startIndex >= len(string):
        raise ValueError("startIndex must be less than length of string")

    if count and (startIndex + count) > len(string):
        raise ValueError("Index and count must refer to a location within the string.")

    return string[:startIndex] + (
        string[startIndex + count :] if count is not None else ""
    )


def replace(string: str, search: str, replace: str):
    return string.replace(search, replace)


def replicate(n: int, x: str) -> str:
    return initialize(n, lambda _=0: x)


def get_char_at_index(input: str, index: int):
    if index < 0 or index >= len(input):
        raise ValueError("Index was outside the bounds of the array.")

    return input[index]


def split(
    string: str,
    splitters: Union[str, List[str]],
    count: Optional[int] = None,
    removeEmpty: int = 0,
) -> List[str]:
    """Split string

    Returns a string array that contains the substrings in this instance
    that are delimited by elements of a specified string or Unicode
    character array."""

    if count and count < 0:
        raise ValueError("Count cannot be less than zero")

    if count == 0:
        return []

    if isinstance(splitters, str):
        if not removeEmpty:
            return string.split(splitters, count - 1 if count else -1)

        splitters = [splitters]

    splitters = [escape(x) for x in splitters] or [" "]

    i = 0
    splits: List[str] = []
    matches = re.finditer("|".join(splitters), string)
    for m in matches:
        if count is not None and count <= 1:
            break

        split = string[i : m.start()]
        if split or not removeEmpty:
            splits.append(split)

            count = count - 1 if count is not None else count

        i = m.end()

    if (count is None or count and count > 0) and len(string) - i > -1:
        split = string[i:]
        if split or not removeEmpty:
            splits.append(split)

    return splits


def trim(string: str, *chars: str) -> str:
    if not len(chars):
        return string.strip()

    pattern = "[" + escape("".join(chars)) + "]+"
    return re.sub(pattern + "$", "", re.sub("^" + pattern, "", string))


def trim_start(string: str, *chars: str) -> str:
    if not len(chars):
        return string.lstrip()

    return re.sub("^[" + escape("".join(chars)) + "]+", "", string)


def trim_end(string: str, *chars: str) -> str:
    if not len(chars):
        return string.rstrip()

    return re.sub("[" + escape("".join(chars)) + "]+$", "", string)


def filter(pred: Callable[[str], bool], x: str) -> str:
    return "".join(c for c in x if pred(c))


def substring(string: str, startIndex: int, length: Optional[int] = None) -> str:
    if startIndex + (length or 0) > len(string):
        raise ValueError("Invalid startIndex and/or length")

    if length is not None:
        return string[startIndex : startIndex + length]

    return string[startIndex:]


class StringComparison(Enum):
    CurrentCulture = 0
    CurrentCultureIgnoreCase = 1
    InvariantCulture = 2
    InvariantCultureIgnoreCase = 3
    Ordinal = 4
    OrdinalIgnoreCase = 5


def cmp(x: str, y: str, ic: Union[bool, StringComparison]) -> int:
    def is_ignore_case(i: Union[bool, StringComparison]) -> bool:
        return (
            i is True
            or i == StringComparison.CurrentCultureIgnoreCase.value
            or i == StringComparison.InvariantCultureIgnoreCase.value
            or i == StringComparison.OrdinalIgnoreCase.value
        )

    def is_ordinal(i: Union[bool, StringComparison]) -> bool:
        return (
            i == StringComparison.Ordinal.value
            or i == StringComparison.OrdinalIgnoreCase.value
        )

    if not x:
        return 0 if not y else -1
    if not y:
        return 1  # everything is bigger than None

    if is_ordinal(ic):
        if is_ignore_case(ic):
            x = x.lower()
            y = y.lower()

        return 0 if x == y else -1 if x < y else 1
    elif is_ignore_case(ic):
        x = x.lower()
        y = y.lower()

    return locale.strcoll(x, y)


@overload
def compare(__string1: str, __string2: str) -> int:
    """Compares two specified String objects and returns an integer that
    indicates their relative position in the sort order."""
    ...


@overload
def compare(
    __string1: str, __string2: str, ignore_case: bool, culture: StringComparison
) -> int:
    ...


def compare(*args: Any) -> int:
    """Compares two specified String objects and returns an integer that
    indicates their relative position in the sort order.

    All overloads of the Compare method return a 32-bit signed integer
    indicating the lexical relationship between the two comparands.

    - Less than zero: The first substring precedes the second
        substring in the sort order.
    - Zero: The substrings occur in the same position in the sort
        order, or length is zero.
    - Greater than zero: The first substring follows the second
        substring in the sort order.

    Returns:
        Integer that indicates the relationship of the two substrings
        to each other in the sort order:
    """
    length = len(args)

    if length == 2:
        return cmp(args[0], args[1], False)
    if length == 3:
        return cmp(args[0], args[1], args[2])
    if length == 4:
        return cmp(args[0], args[1], args[2])
    if length == 5:
        return cmp(
            args[0][args[1] : args[4] + 1], args[2][args[3] : args[4] + 1], False
        )
    if length == 6:
        return cmp(
            args[0][args[1] : args[4] + 1], args[2][args[3] : args[4] + 1], args[5]
        )
    if length == 7:
        return cmp(
            args[0][args[1] : args[4] + 1], args[2][args[3] : args[4] + 1], args[5]
        )
    raise Exception("String.compare: Unsupported number of parameters")


def compare_to(this: str, other: str) -> int:
    """Compare this string with other

    Compares this instance with a specified String object and indicates
    whether this instance precedes, follows, or appears in the same
    position in the sort order as the specified string.
    """
    return cmp(this, other, StringComparison.CurrentCulture)


def ends_with(string: str, search: str):
    idx = string.rfind(search)
    return idx >= 0 and idx == len(string) - len(search)


def starts_with(string: str, pattern: str, ic: int):
    if len(string) >= len(pattern):
        return cmp(string[0 : len(pattern)], pattern, True if ic else False) == 0

    return False


def index_of_any(string: str, any_of: List[str], *args: int):
    if not string:
        return -1

    start_index = args[0] if len(args) > 0 else 0
    if start_index < 0:
        raise ValueError("Start index cannot be negative")

    length = args[1] if len(args) > 1 else len(string) - start_index
    if length < 0:
        raise ValueError("Length cannot be negative")

    if length > len(string) - start_index:
        raise ValueError("Invalid start_index and length")

    string = string[start_index:length]
    for c in any_of:
        index = string.find(c)
        if index > -1:
            return index + start_index

    return -1
